﻿
local me = { name = "target"}
local mod = thismod
mod[me.name] = me

--[[
KTM_MasterTarget.lua

Manages the master target. This is an optional mob to filter our threat on. If it is enabled (<me.mastertarget> is non-nil), only threat towards that target is counted.
]]

me.mastertarget = nil	-- name of the master target, nil for none
me.isworldboss = false	-- whether the master target is a boss. When a boss dies we can clear the master target.

--[[
---------------------------------------------------------------------------------------------
					Slash Commands: Set / Clear the Master Target
---------------------------------------------------------------------------------------------
]]
me.myconsole = 
{
	mastertarget = "sendmastertarget",
}
	
--[[
This function is activated by "/<mod> mastertarget"
]]
me.sendmastertarget = function()
	
	-- check permission
	if mod.net.checkpermission() == nil then
		return
	end
	
	-- do we have a target selected?
	if UnitExists("target") then
		
		-- set target = current
		mod.net.sendmessage("target " .. UnitName("target"))
	
	else
		-- clear target
		mod.net.sendmessage("cleartarget")
	end
	
end

--[[
---------------------------------------------------------------------------------------------
						Loader Services
---------------------------------------------------------------------------------------------
]]

-- Create RelayEvents. 
me.onload = function()
	
	-- Only share Master Target suggestions with players of the same locale
	me.mtsuggest = "mtsuggest" .. mod.string.mylocale
	
	-- Cooldown = 5 secs, max delay = 3 secs.
	me.mtevent = mod.relay.createnew(me.mtsuggest, me.automastertarget, 5.0, 3.0)
	me.bossdeathevent = mod.relay.createnew("bossdeath", me.bossdeath, 5.0, 3.0)
	
end 

--[[
---------------------------------------------------------------------------------------------
						Regex Services
---------------------------------------------------------------------------------------------
]]

me.myparsers = 
{
	{"mobdeath", "UNITDIESOTHER", "CHAT_MSG_COMBAT_HOSTILE_DEATH"}, -- "%s dies."
}

me.onparse = function(identifier, ...)
	
	if identifier == "mobdeath" and me.mastertarget == select(1, ...) and me.isworldboss then
		
		-- Relay the bossdeath event to the raid
		me.bossdeathevent:fireevent()
	end
	
end

--[[
---------------------------------------------------------------------------------------------
					Handling Network Messages
---------------------------------------------------------------------------------------------
]]

me.mynetmessages = { "target", "cleartarget", }

me.onnetmessage = function(author, command, data)
	
	if command == "target" then
		
		-- check the author has permission
		if mod.unit.isplayerofficer(author) == nil then
			return "permission"
		end
		
		-- check newmt makes sense
		if data and string.len(data) > 0 then
			me.setrequest(author, data)
		
		else
			return "invalid"
		end
		
	elseif command == "cleartarget" then
		
		-- check the author has permission
		if mod.unit.isplayerofficer(author) == nil then
			return "permission"
		end
		
		me.clearrequest(author)	
	end
		
end

--[[
me.clearrequest(author)
Called when we receive a network message of someone clearing the master target.

<author>	string; name of the player who sent the message
]]
me.clearrequest = function(author)
	
	-- announce if there is a change
	if me.mastertarget then
		
		me.mastertarget = nil
		mod.raidtable.updateview()
		
		mod.printf(mod.string.get("print", "network", "mtclear"), author)
	end
		
end

--[[
mod.target.setrequest(author, target)
Called when we receive a network message of someone setting the master target.
The problem is that if you have a different localisation to <author>, you will think his target is spelt differently to <target>! So we have to check for this, and override if necessary.

<author>	string; name of the player who sent the message
<target>	string; name of a mob
]]
me.setrequest = function(author, target)
		
	local errormessage = ""	
	
	-- 1) Find the author's UnitID
	local officerunit = mod.unit.findunitidfromname(author)
	local officertarget = UnitName(tostring(officerunit) .. "target")
	
	-- If the officer's target is just far enough away from us, we will be able to target him but his name will be "Unknown", which could stuff things up
	if officertarget == UNKNOWN then
		officertarget = nil
	end
	
	-- 2) Check for differences
	if officertarget == nil then
		errormessage = string.format(mod.string.get("print", "network", "newmttargetnil"), target, author)
	
	elseif officertarget ~= target then
		errormessage = string.format(mod.string.get("print", "network", "newmttargetmismatch"), author, target, officertarget)
		target = officertarget
	end
		
	-- 3) Check for worldboss target
	if UnitClassification((officerunit or "") .. "target") == "worldboss" then
		me.isworldboss = true
	else
		me.isworldboss = false
	end
	
	-- 4) Notify user if there is a change
	if target ~= me.mastertarget then
		me.mastertarget = target
		mod.raidtable.updateview()
		
		-- print out any warning if there was one
		if errormessage ~= "" then
			mod.print(errormessage)
		end
		
		-- print out the new target
		mod.printf(mod.string.get("print", "network", "newmt"), target, author)
	end
		
end

--[[
---------------------------------------------------------------------------------------------
					Relay Event Triggers
---------------------------------------------------------------------------------------------
]]

--[[
This is called when the RelayEvent <me.mtevent> activates, i.e. when a local or remote boss mod sets the master target.
<target> is the localised name of the mob.
]]
me.automastertarget = function(target)

	-- ignore if this is a dupe
	if me.mastertarget == target then
		return
	end

	me.mastertarget = target
	mod.raidtable.updateview()
	
	me.isworldboss = true
	me.lastmttime = GetTime()
	me.lastpollmessage = ""
	
	-- explain to user
	mod.printf(mod.string.get("print", "boss", "automt"), target)
	
end

--[[
This is called when the RelayEvent <me.bossdeathevent> activates, i.e. when a local or remote client reports the death of the master target.
]]
me.bossdeath = function()
	
	me.mastertarget = nil
	mod.raidtable.updateview()
	
	me.isworldboss = false
	me.lastmttime = GetTime()
	me.lastpollmessage = ""
	
end

--[[
---------------------------------------------------------------------------------------------
					Public Methods
---------------------------------------------------------------------------------------------
]]

--[[
mod.target.localsetmt(mtname)

Call this method when the addon sets the master target programmatically, usually from boss mods.

<mtname>	string; name of the master target.
]]
me.localsetmt = function(mtname)
	
	-- fire the mtsuggest RelayEvent
	me.mtevent:fireevent(mtname)
	
end

--[[
mod.target.targetismaster(target)

Checks whether <target> matches the master target, i.e. whether damage against them will count against your threat total.

<target>	string; the name of the mob being queried.
Returns: non-nil if <target> matches the mastertarget.
]]
me.targetismaster = function(target)
	
	if me.mastertarget == nil then
		return true
	end
	
	if target == me.mastertarget then
		return true
	end
	
	-- insert other checks here, later.
	
	return -- (nil)
	
end
